mod model;
use model::Model;

mod scripting;
use scripting::{ ScriptLang, Lua };

mod gameobject;
use gameobject::*;

#[macro_use]
extern crate glium;
extern crate nalgebra_glm as glm;

use std::rc::Rc;
use std::cell::RefCell;

fn main() {
    use glium::{glutin, Surface};
    
    let events_loop = glutin::event_loop::EventLoop::new();
    let wb = glutin::window::WindowBuilder::new()
        .with_title("Couch Game Engine");
    let cb = glutin::ContextBuilder::new()
	.with_depth_buffer(24);

    let display = glium::Display::new(wb, cb, &events_loop).unwrap();

    let vertex_shader_src = include_str!("flat.vert");
    let fragment_shader_src = include_str!("flat.frag");
    let program = glium::Program::from_source(&display, vertex_shader_src, fragment_shader_src, None).unwrap();

    let root = Rc::new(RefCell::new(Node::new("_root_".to_string(), GameObject::Null)));
    {
	// Initial scene setup
	let mut root = root.borrow_mut();
	
	let mut cube = Mesh::new(Model::new(&display, "Box.glb", "container.png"));
	cube.set_transform(gameobject::Transform {
	    position: glm::vec3(0.0, 0.0, -10.0),
	    rotation: glm::vec3(0.0, 15.0_f32.to_radians(), 0.0),
	    scale: glm::vec3(1.0, 1.0, 1.0),
	});
	root.add_child(Rc::new(RefCell::new(Node::new("cube".to_string(), GameObject::Mesh(Box::new(cube))))));
    }
	
    let mut script_lang = Lua::new(Rc::clone(&root));
    script_lang.init();

    let mut last_draw = std::time::Instant::now();

    events_loop.run(move |ev, _, control_flow| {
        let next_frame_time = std::time::Instant::now() +
            std::time::Duration::from_nanos(16_666_667);
        *control_flow = glutin::event_loop::ControlFlow::WaitUntil(next_frame_time);
	match ev {
	    glutin::event::Event::WindowEvent { event, .. } => {
		if event == glutin::event::WindowEvent::CloseRequested {
                    *control_flow = glutin::event_loop::ControlFlow::Exit;
		}
	    },
	    glutin::event::Event::DeviceEvent { event, .. } => {
		if let glutin::event::DeviceEvent::Key(
		    glutin::event::KeyboardInput { scancode, state, .. }
		) = event {
		    match state {
			glutin::event::ElementState::Pressed => script_lang.key_pressed(scancode),
			glutin::event::ElementState::Released => script_lang.key_released(scancode),
		    }
		};
	    },
	    glutin::event::Event::MainEventsCleared => {
		let now = std::time::Instant::now();
		let delta = (now - last_draw).as_secs_f32();
		last_draw = now;

		script_lang.update(delta);

		let mut target = display.draw();
		let params = glium::DrawParameters {
		    depth: glium::Depth {
			test: glium::DepthTest::IfLess,
			write: true,
			.. Default::default()
		    },
		    backface_culling: glium::draw_parameters::BackfaceCullingMode::CullClockwise,
		    .. Default::default()
		};
		target.clear_color_and_depth((0.0, 0.0, 0.0, 1.0), 1.0);

		let root = root.borrow();

		let cube = root.find_node("cube").unwrap();
		let mut cube = cube.borrow_mut();
		let cube: &mut Mesh = match &mut cube.gameobject {
		    GameObject::Mesh(mesh) => mesh,
		    _ => panic!("cube is not a mesh"),
		};
		// TODO, do this over the whole scene tree
		let mvp = glm::perspective::<f32>(4.0/3.0, 45.0_f32.to_radians(), 0.2, 100.0);
		let mvp = glm::translate::<f32>(&mvp, &cube.get_transform().position);
		let mvp = glm::rotate_x(&mvp, cube.get_transform().rotation.x);
		let mvp = glm::rotate_y(&mvp, cube.get_transform().rotation.y);
		let mvp = glm::rotate_z(&mvp, cube.get_transform().rotation.z);
		
		cube.model.draw(&mut target, &program, &mvp, &params);

		target.finish().unwrap();
	    }
	    _ => (),
        };
    });
}
